﻿using UnityEngine;
using System.Collections.Generic;
using SimpleJSON;

namespace MacGruber
{
    public class RandomChoice : MVRScript
    {
		private JSONStorableStringChooser myChoiceChooser;
		private JSONStorableBool myAvoidRepetition;
		private JSONStorableString myChoiceName;
		private JSONStorableFloat myChoiceProbability;

		private List<object> myChoiceUI = new List<object>();

		private class Choice
		{
			public Choice(MVRScript script, string name)
			{
				myName = name;
				Trigger = new EventTrigger(script, "On" + name);
			}

			public static int SortByNameAscending(Choice a, Choice b)
			{
				return a.Name.CompareTo(b.Name);
			}

			public string Name {
				get { return myName; }
				set {
					myName = value;
					Trigger.Name = "On" + value;
				}
			}

			public float Probability = 0.5f;
			public EventTrigger Trigger;
			private string myName;
		}

		private List<Choice> myChoices = new List<Choice>();
		private int myPreviousIndex1 = -1;
		private int myPreviousIndex2 = -1;
		private bool myNeedUIRefresh = false;

        public override void Init()
		{
			Utils.OnInitUI(CreateUIElement);

			Utils.SetupInfoText(this,
				"<color=#606060><size=40><b>RandomSelect</b></size>\nLogicBrick that, after <i>Trigger</i> is called, selects a random <i>Choice</i> to trigger.\n\n" +
				"Use <i>Avoid Repetition</i> take the last one or two choices out of the selection, depending on the number of total available choices.</color>",
				300.0f, true
			);

			myChoiceChooser = Utils.SetupStringChooser(this, "Choice", new List<string>(), false);
			myChoiceChooser.isStorable = myChoiceChooser.isRestorable = false;
			myChoiceChooser.setCallbackFunction += UISetChoice;

			Utils.SetupButton(this, "Add Choice", UIAddChoice, false);
			Utils.SetupButton(this, "Remove Choice", UIRemoveChoice, false);
			myAvoidRepetition = Utils.SetupToggle(this, "Avoid Repetition", true, false);

			Utils.SetupButton(this, "Test Trigger", TriggerAction, false);
			Utils.SetupAction(this, "Trigger", TriggerAction);

			Utils.SetupSpacer(this, 15, false);

			myChoiceName = new JSONStorableString("Name", "", UIRenameChoice);
			myChoiceProbability = new JSONStorableFloat("Relative Probability", 0.5f, 0.0f, 1.0f, true, true);
			myChoiceProbability.setCallbackFunction += UISetProbability;

			SuperController.singleton.onAtomUIDRenameHandlers += OnAtomRename;
			SimpleTriggerHandler.LoadAssets();
		}

		private void UISetChoice(string v)
		{
			Utils.RemoveUIElements(this, myChoiceUI);

			int idx = myChoiceChooser.choices.IndexOf(v);
			if (idx < 0)
				return;

			Choice choice = myChoices[idx];

			JSONStorableString optionHeader = Utils.SetupInfoOneLine(this, "<size=30><b>Choice</b></size>", false);
			myChoiceUI.Add(optionHeader);

			myChoiceName.valNoCallback = choice.Name;
			UIDynamicLabelInput nameInput = Utils.SetupTextInput(this, "Name", myChoiceName, false);
			myChoiceUI.Add(nameInput);

			myChoiceProbability.valNoCallback = choice.Probability;
			CreateSlider(myChoiceProbability, false);
			myChoiceUI.Add(myChoiceProbability);

			UIDynamicButton actionsButton = Utils.SetupButton(this, "Actions " + choice.Trigger.Name, choice.Trigger.OpenPanel, false);
			myChoiceUI.Add(actionsButton);
		}

		private void UISetProbability(float v)
		{
			int idx = myChoiceChooser.choices.IndexOf(myChoiceChooser.val);
			if (idx < 0)
				return;

			myChoices[idx].Probability = v;
		}

		private void UIAddChoice()
		{
			for (int i=1; i<1000; ++i)
			{
				string name = "Choice" + i;
				if (!myChoiceChooser.choices.Contains(name))
				{
					myChoices.Add(new Choice(this, name));
					myChoices.Sort(Choice.SortByNameAscending);
					List<string> options = myChoiceChooser.choices;
					options.Clear();
					for (int e=0; e<myChoices.Count; ++e)
						options.Add(myChoices[e].Name);
					myChoiceChooser.choices = null; // force UI sync
					myChoiceChooser.choices = options;
					myChoiceChooser.val = name;
					return;
				}
			}
			SuperController.LogError("RandomChoice: Too many options!");
		}

		private void UIRemoveChoice()
		{
			int idx = myChoiceChooser.choices.IndexOf(myChoiceChooser.val);
			if (idx < 0)
				return;

			List<string> options = myChoiceChooser.choices;
			options.RemoveAt(idx);
			myChoices[idx].Trigger.Remove();
			myChoices.RemoveAt(idx);
			myChoiceChooser.choices = null; // force UI sync
			myChoiceChooser.choices = options;

			if (idx < options.Count)
				myChoiceChooser.val = options[idx];
			else if (options.Count > 0)
				myChoiceChooser.val = options[options.Count-1];
			else
				myChoiceChooser.val = "";
		}

		private void UIRenameChoice(string name)
		{
			int idx = myChoiceChooser.choices.IndexOf(myChoiceChooser.val);
			if (idx < 0)
				return;

			myChoices[idx].Name = name;

			myChoices.Sort(Choice.SortByNameAscending);
			List<string> options = myChoiceChooser.choices;
			options.Clear();
			for (int e=0; e<myChoices.Count; ++e)
				options.Add(myChoices[e].Name);
			myChoiceChooser.choices = null; // force UI sync
			myChoiceChooser.choices = options;
			myChoiceChooser.val = name;
		}

		private void OnAtomRename(string oldid, string newid)
		{
			for (int i=0; i<myChoices.Count; ++i)
				myChoices[i].Trigger.SyncAtomNames();
		}

		public override JSONClass GetJSON(bool includePhysical = true, bool includeAppearance = true, bool forceStore = false)
		{
			JSONClass jc = base.GetJSON(includePhysical, includeAppearance, forceStore);
			if (includePhysical || forceStore)
			{
				needsStore = true;

				JSONArray clist = new JSONArray();
				for (int i=0; i<myChoices.Count; ++i)
				{
					Choice choice = myChoices[i];
					JSONClass cc = new JSONClass();
					cc["Name"] = choice.Name;
					cc["Probability"].AsFloat = choice.Probability;
					cc[choice.Trigger.Name] = choice.Trigger.GetJSON(base.subScenePrefix);
					clist.Add("", cc);
				}
				jc["Choices"] = clist;
			}

			return jc;
		}

		public override void LateRestoreFromJSON(JSONClass jc, bool restorePhysical = true, bool restoreAppearance = true, bool setMissingToDefault = true)
		{
			base.LateRestoreFromJSON(jc, restorePhysical, restoreAppearance, setMissingToDefault);
			if (!base.physicalLocked && restorePhysical && !IsCustomPhysicalParamLocked("trigger"))
			{
				for (int i=0; i<myChoices.Count; ++i)
					myChoices[i].Trigger.Remove();
				myChoices.Clear();

				JSONArray clist = jc["Choices"].AsArray;
				for (int i=0; i<clist.Count; ++i)
				{
					JSONClass cc = clist[i].AsObject;
					Choice choice = new Choice(this, cc["Name"]);
					choice.Probability = cc["Probability"].AsFloat;
					choice.Trigger.RestoreFromJSON(cc, base.subScenePrefix, base.mergeRestore, setMissingToDefault);
					myChoices.Add(choice);
				}

				myChoices.Sort(Choice.SortByNameAscending);
				List<string> options = myChoiceChooser.choices;
				options.Clear();
				for (int e=0; e<myChoices.Count; ++e)
					options.Add(myChoices[e].Name);
				myChoiceChooser.choices = null; // force UI sync
				myChoiceChooser.choices = options;

				if (options.Count == 0)
					myChoiceChooser.valNoCallback = "";
				else if (!options.Contains(myChoiceChooser.val))
					myChoiceChooser.valNoCallback = options[0];
				myNeedUIRefresh = true;
			}
		}

		private void Update()
		{
			for (int i=0; i<myChoices.Count; ++i)
				myChoices[i].Trigger.Update();
			
			if (myNeedUIRefresh)
			{
				myNeedUIRefresh = false;
				myChoiceChooser.setCallbackFunction(myChoiceChooser.val);
			}
		}

		private void OnDestroy()
		{
			SuperController.singleton.onAtomUIDRenameHandlers -= OnAtomRename;

			for (int i=0; i<myChoices.Count; ++i)
				myChoices[i].Trigger.Remove();
			myChoices.Clear();

			Utils.OnDestroyUI();
		}

		private void TriggerAction()
		{
			if (myChoices.Count == 0)
				return;

			bool avoidRepetition = myAvoidRepetition.val;
			if (myChoices.Count < 3 || myPreviousIndex1 >= myChoices.Count)
				myPreviousIndex1 = -1;
			if (myChoices.Count < 4 || myPreviousIndex2 >= myChoices.Count)
				myPreviousIndex2 = -1;

			float sum = 0.0f;
			for (int i=0; i<myChoices.Count; ++i)
			{
				if (avoidRepetition && (i == myPreviousIndex1 || i == myPreviousIndex2))
					continue;

				sum += myChoices[i].Probability;
			}
			if (sum == 0.0f)
				return;

			float threshold = UnityEngine.Random.Range(0.0f, sum);
			sum = 0.0f;
			int idx = -1;
			for (int i=0; i<myChoices.Count; ++i)
			{
				if (avoidRepetition && (i == myPreviousIndex1 || i == myPreviousIndex2))
					continue;
				idx = i;
				sum += myChoices[i].Probability;
				if (threshold <= sum)
					break;
			}

			myPreviousIndex2 = myPreviousIndex1;
			myPreviousIndex1 = idx;
			if (idx >= 0)
				myChoices[idx].Trigger.Trigger();
		}
    }
}
